#include "ANSim2omnet.h"

void ANSim2omnet::parseXMLFile(const char* filename) {

  doc = xmlParseFile(filename);
  cur = xmlDocGetRootElement(doc); // TEST
  ns = xmlSearchNsByHref(doc, cur, (const xmlChar *) "http://www.gnome.org/some-location");

  if (xmlStrcmp(cur->name, (const xmlChar *) "simulation")) {
    fprintf(stderr,"document of the wrong type, root node != Helping");
    xmlFreeDoc(doc);
    return;
  }

  // Initialize countID everytime a new XML file is parsed
  countID = 0;

  while ( cur && xmlIsBlankNode ( cur ) ) {
    cur = cur -> next;
  }

  
  /* Second level is the list of  parameters */
  cur = cur->xmlChildrenNode;
  while (cur != NULL) {
    if ((!xmlStrcmp(cur->name, (const xmlChar *) "parameter")) && (cur->ns == ns)) {
      parameter = parseParameter(doc, ns, cur);
 
    }
    if ((!xmlStrcmp(cur->name, (const xmlChar *) "node_settings")) && (cur->ns == ns)) {
      nodeList = parseNodeSettings(doc, ns, cur, parameter.numberOfNodes);
#ifdef DEBUG
      for(i=0; i<parameter.numberOfNodes; i++) {
	printf("NodeID: %d xpos: %lf ypos: %lf\n", nodeList[i].node_id, nodeList[i].position.xpos, nodeList[i].position.ypos);
	
      }
#endif
 
    }
    if ((!xmlStrcmp(cur->name, (const xmlChar *) "mobility")) && (cur->ns == ns)) {
      parseMobility(doc, ns, cur);
    }
    if ((!xmlStrcmp(cur->name, (const xmlChar *) "statistics")) && (cur->ns == ns)) {
      xmlNodePtr curTemp = cur->xmlChildrenNode;
      while (curTemp!=NULL) { 
	if ((!xmlStrcmp(curTemp->name, (const xmlChar *) "stoptime")) && (curTemp->ns == ns)) {
	  stopTime = atof((char *)xmlNodeListGetString(doc, curTemp->xmlChildrenNode, 1));
	}
	curTemp = curTemp->next;
      }
      
    }
    cur = cur->next;
  }
  
}

Parameter ANSim2omnet::parseParameter(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur) {

  Parameter ret;

  cur = cur->xmlChildrenNode;
    while (cur != NULL) {
        
        if ((!xmlStrcmp(cur->name, (const xmlChar *) "area_shape")) &&
	    (cur->ns == ns)) {
	    ret.area_shape = (char *)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);

	}
        if ((!xmlStrcmp(cur->name, (const xmlChar *) "xsize")) &&
	    (cur->ns == ns)) {
	    ret.xsize = atoi((char *)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1));
	}
        if ((!xmlStrcmp(cur->name, (const xmlChar *) "ysize")) &&
	    (cur->ns == ns)) {
	    ret.ysize = atoi((char *)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1));
	}
        if ((!xmlStrcmp(cur->name, (const xmlChar *) "extension_factor")) &&
	    (cur->ns == ns)) {
	    ret.extension_factor = atoi((char *)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1));
	}
        if ((!xmlStrcmp(cur->name, (const xmlChar *) "range")) &&
	    (cur->ns == ns)) {
	    ret.range = atof((char *)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1));
	}

        if ((!xmlStrcmp(cur->name, (const xmlChar *) "mobility_model")) &&
	    (cur->ns == ns)) {
	    ret.mobility_model = (char *)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
	}

        if ((!xmlStrcmp(cur->name, (const xmlChar *) "max_velocity")) &&
	    (cur->ns == ns)) {
	    ret.max_velocity = atof((char *)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1));
	}
        if ((!xmlStrcmp(cur->name, (const xmlChar *) "max_stopTime")) &&
	    (cur->ns == ns)) {
	    ret.max_stopTime = atof((char *)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1));
	}
        if ((!xmlStrcmp(cur->name, (const xmlChar *) "max_moveTime")) &&
	    (cur->ns == ns)) {
	    ret.max_moveTime = atof((char *)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1));
	}

	if ((!xmlStrcmp(cur->name, (const xmlChar *) "numberOfNodes")) &&
	    (cur->ns == ns)) {
	    ret.numberOfNodes = atoi((char *)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1));
	}        
        if ((!xmlStrcmp(cur->name, (const xmlChar *) "communication_mode")) &&
	    (cur->ns == ns)) {
	    ret.communication_mode = (char *)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
	}

	cur = cur->next;
    }

    return(ret);

}



Node* ANSim2omnet::parseNodeSettings (xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur, int numberOfNodes) {

  Node* nodeList;
  int i;
  xmlNodePtr curNode, curPosition;

  nodeList = new Node[numberOfNodes];

  cur=cur->xmlChildrenNode;  
  i = 0;

  while(cur!=NULL) { 
    if ((!xmlStrcmp(cur->name, (const xmlChar *) "node")) && (cur->ns == ns)) {
      curNode=cur->xmlChildrenNode;
      while(curNode!=NULL) {
	if ((!xmlStrcmp(curNode->name, (const xmlChar *) "node_id")) && (curNode->ns == ns)) {
	  nodeList[i].node_id= atoi((char *)xmlNodeListGetString(doc, curNode->xmlChildrenNode, 1));
	
	}
	if ((!xmlStrcmp(curNode->name, (const xmlChar *) "position")) && (curNode->ns == ns)) {
	  curPosition=curNode->xmlChildrenNode;
	  while(curPosition!=NULL) {
	    if ((!xmlStrcmp(curPosition->name, (const xmlChar *) "xpos")) && (curPosition->ns == ns)) {
	  nodeList[i].position.xpos= atof((char *)xmlNodeListGetString(doc, curPosition->xmlChildrenNode, 1));
	    }
	    if ((!xmlStrcmp(curPosition->name, (const xmlChar *) "ypos")) && (curPosition->ns == ns)) {
	      nodeList[i].position.ypos= atof((char *)xmlNodeListGetString(doc, curPosition->xmlChildrenNode, 1));
	    }	    
	    curPosition = curPosition->next;
	  }	
	  
	}	  

	curNode = curNode->next;
      }
      i++;
    }
    cur=cur->next;
  }
  
  return nodeList;
  
}

void ANSim2omnet::parseMobility(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur) {

  xmlNodePtr curPositionChange, curDestination;
  int node_id;  // OKKIO !!!
  PositionChange positionChange;

  cur=cur->xmlChildrenNode;  

  while(cur!=NULL) { 
    if ((!xmlStrcmp(cur->name, (const xmlChar *) "position_change")) && (cur->ns == ns)) {
      curPositionChange=cur->xmlChildrenNode;
      while(curPositionChange!=NULL) {
	if ((!xmlStrcmp(curPositionChange->name, (const xmlChar *) "node_id")) && (curPositionChange->ns == ns)) {
	  node_id= atoi((char *)xmlNodeListGetString(doc, curPositionChange->xmlChildrenNode, 1));
	
	}
	if ((!xmlStrcmp(curPositionChange->name, (const xmlChar *) "start_time")) && (curPositionChange->ns == ns)) {
	  positionChange.start_time= atof((char *)xmlNodeListGetString(doc, curPositionChange->xmlChildrenNode, 1));
	
	}
	if ((!xmlStrcmp(curPositionChange->name, (const xmlChar *) "velocity")) && (curPositionChange->ns == ns)) {
	  positionChange.velocity= atof((char *)xmlNodeListGetString(doc, curPositionChange->xmlChildrenNode, 1));
	
	}
	if ((!xmlStrcmp(curPositionChange->name, (const xmlChar *) "destination")) && (curPositionChange->ns == ns)) {
	  curDestination=curPositionChange->xmlChildrenNode;
	  while(curDestination!=NULL) {
	    if ((!xmlStrcmp(curDestination->name, (const xmlChar *) "xpos")) && (curDestination->ns == ns)) {
	      positionChange.destination.xpos= atof((char *)xmlNodeListGetString(doc, curDestination->xmlChildrenNode, 1));
	    }
	    if ((!xmlStrcmp(curDestination->name, (const xmlChar *) "ypos")) && (curDestination->ns == ns)) {
	      positionChange.destination.ypos= atof((char *)xmlNodeListGetString(doc, curDestination->xmlChildrenNode, 1));
	    }	    
	    curDestination = curDestination->next;
	  }	
	  
	}	  

	curPositionChange = curPositionChange->next;
      }
      nodeList[node_id].mobility.push(positionChange);
    }
    cur=cur->next;
  }
}


ANSim2omnet::~ANSim2omnet() {

  delete[] nodeList;
}
