#include <stdio.h>
#include <string.h>
#include <math.h>
#include <iostream>
#include <fstream>
#include <coutvector.h>
#include "omnetpp.h"
#include "dataCollector.h"
#include "mobMessage_m.h"
#include "definitions.h"

using namespace std;

Define_Module(DataCollector);



void DataCollector::initialize()
{
	//initialization

	//retrieving values of the parameters
	cModule* parent=getParentModule();


	m_iNumberOfMessages=(int)parent->par("numberOfMessages").longValue();
	m_iNumberOfBlackholeCarriers = parent->par("numberOfActiveNonCollaborativeHosts");
	m_iNumberOfCarriers = parent->par("numberOfCarriers");
	
	m_iNumberOfGoodMessagesReceived = 0;
	m_iNumberOfAckReceived = 0;
	m_iNumberOfMessagesReceived = 0;
	m_iNumberOfBlackholeMessages = 0;
	m_iNumberOfGoodMessages = 0;
	m_dPartialDelay = 0.0;
	m_iTotalNumberOfSyncMessagesSent = 0;
	m_iTotalNumberOfAsyncMessagesSent = 0;
	m_iTotalNumberOfRoutingMessagesSent = 0;
	m_iNumberOfAsyncMessagesReceived = 0;
	m_iNumberOfAttractedMessages = 0;
	m_iNumberOfForwardedMessages = 0;
	m_iNumberOfPenalizedMessages = 0;
	m_iNumberOfAsyncMessagesSent = 0;
	m_iNumberOfBadMessagesReceived = 0;

	m_NumberOfAttractedMsg.setName("Number of Attracted Msg");
	m_NumberOfDeliveredMsg.setName("Number of Delivered Msg");
	m_NumberOfSentMsg.setName("Number of Sent Msg");

	int i = 0;
	for (i = 0; i < m_iNumberOfMessages; i++)
		m_bMsgIdBuffer[i] = false;

	cMessage* msg = new cMessage("dataCollector");
	m_iCollectTimer = 500;
	scheduleAt(simTime()+ m_iCollectTimer ,msg);
}

void DataCollector::vIncrementNumberOfMessageAttracted(int iMsgId)
{
	if(! m_bMsgIdBuffer[iMsgId])
	{
		m_bMsgIdBuffer[iMsgId] = true;
		m_iNumberOfAttractedMessages++;
	}
}


void DataCollector::vIncrementNumberOfMessagesReceived(int iHostId, double dDelay)
{

	m_iNumberOfMessagesReceived++;

	if ( ( iHostId < m_iNumberOfBlackholeCarriers ) ) 
	{
		m_iNumberOfBadMessagesReceived++;
		return;
	}


	m_iNumberOfGoodMessagesReceived++;
	m_dPartialDelay += dDelay;	


}

void DataCollector::handleMessage(cMessage *msg)
{
	m_NumberOfAttractedMsg.record(m_iNumberOfAttractedMessages);
	m_NumberOfDeliveredMsg.record(m_iNumberOfGoodMessagesReceived);
	m_NumberOfSentMsg.record(m_iNumberOfGoodMessages);

	scheduleAt(simTime()+ m_iCollectTimer ,msg);
}

void DataCollector::finish(){

	//double iNumberOfBadMessagesReceived = double(m_iNumberOfMessagesReceived - m_iNumberOfGoodMessagesReceived);
	double deliveryRatio =  double(m_iNumberOfGoodMessagesReceived) /double(m_iNumberOfGoodMessages);
	double blackholeDeliveryRatio =  double(m_iNumberOfBadMessagesReceived) /double(m_iNumberOfBlackholeMessages);
	//double dAsyncDeliveryRatio =  double(m_iNumberOfAsyncMessagesReceived) /double(m_iNumberOfAsyncMessagesSent);
	recordScalar("Number of Good Messages received", double(m_iNumberOfGoodMessagesReceived));
	recordScalar("Number of Messages received", double(m_iNumberOfMessagesReceived));
	recordScalar("Number of Bad Messages received", m_iNumberOfBadMessagesReceived );
	recordScalar("Number of Blackhole Messages sent", double(m_iNumberOfBlackholeMessages));
	recordScalar("Number of Good Messages sent", double(m_iNumberOfGoodMessages));
	recordScalar("Number of Async Messages received", double(m_iNumberOfAsyncMessagesReceived));
	recordScalar("Attraction ratio", double(m_iNumberOfAttractedMessages)/double(m_iNumberOfMessages));
	recordScalar("Number of Attracted messages",double(m_iNumberOfAttractedMessages));
	recordScalar("Number of Forwarded Messages", double(m_iNumberOfForwardedMessages));
	recordScalar("Number of Penalized Messages", double(m_iNumberOfPenalizedMessages));
	recordScalar("Blackhole Delivery ratio", blackholeDeliveryRatio);
	recordScalar("Delivery ratio", deliveryRatio);
	recordScalar("Total Number of Sent Messages", m_iTotalNumberOfSyncMessagesSent + m_iTotalNumberOfAsyncMessagesSent + m_iTotalNumberOfRoutingMessagesSent);
	recordScalar("Total Number of Sync Sent Messages", m_iTotalNumberOfSyncMessagesSent);
	recordScalar("Total Number of Async Sent Messages", m_iTotalNumberOfAsyncMessagesSent);
	recordScalar("Total Number of Routing Sent Messages", m_iTotalNumberOfRoutingMessagesSent);	
	recordScalar("Average Delay", m_dPartialDelay/double(m_iNumberOfGoodMessagesReceived));
	recordScalar("Number of Ack Received", m_iNumberOfAckReceived);
	//recordScalar("Blackhole Delivery Ratio", double(m_iNumberOfBadMessagesReceived/m_iNumberOfBlackholeMessages));

}

