#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <fstream>
#include <iostream>
#include <cstdlib>
#include <ios>
#include "omnetpp.h"
#include "engine.h"
#include "host.h"
#include "definitions.h"

Define_Module(Engine);

void Engine::initialize()
{	
	//retrieving values of the parameters
	cModule* parent=getParentModule();
	numHosts=parent->par("numberOfHosts");
								   
	
	// assign hosts to groups in the same way of movGen (program for traces generation)
	for(int i = 0; i < numHosts; i++)
	{
		int groupId=i%numberOfGroups;

		Host *HostModule = (Host*)(parent->getSubmodule("Hosts", i));
		HostModule->m_dStartingGroup = groupId+1;
	}
	
	string mobilityTracesFileStr = parent->par("socMovFile");
	const char* mobilityTracesFile = mobilityTracesFileStr.c_str();
	
	//Parsing XML File
	ANSim2omnet *parser = new ANSim2omnet();
	parser->parseXMLFile(mobilityTracesFile);
	nodeList=parser->getNodeList();
	scenario=parser->getParameter();
	range=scenario.range;
	
	//Initialization of the initial positions of the nodes
	
    	initPositions(); 	
	cMessage* startMessage = new cMessage("Start",RECONFIGURATION);
	scheduleAt(0.00, startMessage);	
}

void Engine::handleMessage(cMessage* msg)
{
	delete msg;
	updatePositions();
	
	cMessage* startMessage = new cMessage("Start",RECONFIGURATION);
	scheduleAt(simTime()+1.0, startMessage);	
}

void Engine::initPositions()
{	
	cModule* parent=getParentModule();
	
	for (int i=0; i<numHosts; i++)
	{	
		//extracting positions
		positions[i] = nodeList[i].position;
		hosts[i].currentX=(int)positions[i].xpos+scenario.xsize/2;
		hosts[i].currentY=(int)positions[i].ypos+scenario.ysize/2;

		//updating current position...
		Host *HostModule=(Host*)(parent->getSubmodule("Hosts",i));
		cDisplayString dispstr("i=pda1;b=25,36");	
		char stringcurrentx[20];
		char stringcurrenty[20];	
		sprintf(stringcurrentx,"%d",hosts[i].currentX);
		sprintf(stringcurrenty,"%d",hosts[i].currentY);
		dispstr.insertTag("p");
		dispstr.setTagArg("p",0,stringcurrentx);
		dispstr.setTagArg("p",1,stringcurrenty);
		HostModule->setDisplayString(dispstr.str());
	}
}


void Engine:: updatePositions()
{
	cModule* parent=getParentModule();
	
	double nextMoveTime;
	double currentTime=SIMTIME_DBL(simTime());
	PositionChange positionChange;
	double cos, sin, pathLength, diffX, diffY;

	for (int i=0; i<numHosts; i++)
	{	
		if (nodeList[i].mobility.size()>1)
		{
			nextMoveTime=getNextMoveTime(i);

			if (nextMoveTime<currentTime)
			{
				nodeList[i].mobility.pop();
			}
			positionChange=nodeList[i].mobility.front();
			diffX=positionChange.destination.xpos-positions[i].xpos;
			diffY=positionChange.destination.ypos-positions[i].ypos;
			pathLength= pow(pow(diffX,2)+pow(diffY,2),0.5);
			cos=diffX/pathLength;
			sin=diffY/pathLength;
			positions[i].xpos=positionChange.velocity*1.0*cos+positions[i].xpos;
			positions[i].ypos=positionChange.velocity*1.0*sin+positions[i].ypos;
		
			hosts[i].currentX=(int)positions[i].xpos+scenario.xsize/2;
			hosts[i].currentY=(int)positions[i].ypos+scenario.ysize/2;
			
			//updating current position...
			Host *HostModule=(Host*)(parent->getSubmodule("Hosts",i));
			cDisplayString dispstr("i=pda1;b=25,36");	
			char stringcurrentx[20];
			char stringcurrenty[20];	
			sprintf(stringcurrentx,"%d",hosts[i].currentX);
			sprintf(stringcurrenty,"%d",hosts[i].currentY);
			dispstr.insertTag("p");
			dispstr.setTagArg("p",0,stringcurrentx);
			dispstr.setTagArg("p",1,stringcurrenty);
			HostModule->setDisplayString(dispstr.str());
		}
	}
	
	//Setting connectivity
	
	for (int i=0;i<numHosts;i++)
	{
		//update connectivity
		for (int j=0;j<numHosts;j++)
		{

			Host *HostModule_i=(Host *)parent->getSubmodule("Hosts",i);
			Host *HostModule_j=(Host *)parent->getSubmodule("Hosts",j);
			
			if ((HostModule_i->gate("outputs",j))->isConnected())
				HostModule_i->gate("outputs",j)->disconnect();
	
			if ((HostModule_j->gate("inputs",i))->isConnected()) 
				HostModule_j->gate("inputs",i)->disconnect();
		}
	}
	
	for (int i=0;i<numHosts;i++)
	{
		//update connectivity
		for (int j=0;j<numHosts;j++)
		{
			
			Host *HostModule_i=(Host *)parent->getSubmodule("Hosts",i);
			Host *HostModule_j=(Host *)parent->getSubmodule("Hosts",j);
		
			//calculation of the current distance
			double currentDistance=sqrt((double)(hosts[i].currentX-hosts[j].currentX)*(hosts[i].currentX-hosts[j].currentX)+(hosts[i].currentY-hosts[j].currentY)*(hosts[i].currentY-hosts[j].currentY));
			
			
			//if currentDistance<=radius then the hosts are connected
			if (currentDistance<radius)
			{				
				//if the hosts has been previously disconnected, then they must be connected
				HostModule_i->gate("outputs",j)->connectTo(HostModule_j->gate("inputs",i));
				HostModule_i->gate("outputs",j)->setDisplayString("m=a;o=black,1");
				
				//setting connections	
				isConnected[i][j]=true;
					
				//setting reachability
				HostModule_i->setIsInReach(j,true);
				HostModule_j->setIsInReach(i,true);					
				
			}
			//else they are disconnected
			else
			{
				//if the hosts have been previously connected, then they must be disconnected
				(HostModule_i->gate("outputs",j))->connectTo(HostModule_i->gate("inputs",j));
				HostModule_i->gate("outputs",j)->setDisplayString("m=a,0,0,0,0;o=black,1");
				
				isConnected[i][j]=false;
				
				//setting reachability
				HostModule_i->setIsInReach(j,false);
				HostModule_j->setIsInReach(i,false);
				
			}//end else 			
		}
	}
	//end setting connectivity
}

double Engine::getNextMoveTime(int nodeId)
{		
	PositionChange positionChange=nodeList[nodeId].mobility.front();
	return positionChange.start_time;
}


