#!/bin/bash

output="scenarioA.xml"
xsize="1000.000000"
ysize="1000.000000"
numberOfNodes="46"
numberOfCarriers="30"
numberOfHosts=$(echo "$numberOfNodes - $numberOfCarriers" | bc)
range="200.000000"
numberOfLines="4"
simuTime="5000"

if [ -f $output ];
then
	rm $output
fi

echo "<?xml version=\"1.0\"?>
<simulation xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"
xsi:noNamespaceSchemaLocation=\"http://www.i-u.de/schools/hellbrueck/ansim/xml/simulation.xsd\">
	<parameter>
		<area_shape>rectangle</area_shape>
		<xsize>$xsize</xsize>
		<ysize>$ysize</ysize>
		<extension_factor>1</extension_factor>
		<numberOfNodes>$numberOfNodes</numberOfNodes>
		<range>$range</range>
		<mobility_model>socialFoundedMM</mobility_model>
	</parameter>" > $output

numberOfCarriersPerLine=$(echo "$numberOfCarriers / $numberOfLines" | bc)

echo "<node_settings>" >> $output

counter=0
currentLine=0
xmidsize=$(echo "$xsize / 2" | bc)
ymidsize=$(echo "$ysize / 2" | bc)
x0=$(echo "$xmidsize - 100" | bc -l)
y0=$(echo "$ymidsize - 100" | bc -l )

while [ "$counter" -lt "$numberOfCarriers" ];
do
	 	if [ "$currentLine" = "0" ];
		then
			xpos0=$x0
			ypos0=$y0
		fi
		if [ "$currentLine" = "1" ];
		then
			xpos0="-$x0"
			ypos0=$y0
		fi
		if [ "$currentLine" = "2" ];
		then
			xpos0="-$x0"
			ypos0="-$y0"
		fi
		if [ "$currentLine" = "3" ];
		then
			xpos0=$x0
			ypos0="-$y0"
		fi
							
		echo	"	<node>
				<node_id>$counter</node_id>
				<position>
					<xpos>$xpos0.000000</xpos>
					<ypos>$ypos0.000000</ypos>
				</position>
			</node>" >> $output
		xpos[$counter]=$xpos0
		ypos[$counter]=$ypos0
		xy[$counter]="x"
		if [ "${xpos[$counter]}" = "-400" ];
		then
			versus[$counter]="p"
		else
			versus[$counter]="n"
		fi
		RANGE=20
		speed=$RANDOM
		let speed%=$RANGE
		speed[$counter]=$speed
		#echo ${speed[$counter]}

		let counter=counter+1
		let current=current+1
		
		if [ "$current" = "$numberOfCarriersPerLine" ];
		then
			current=0
			let currentLine=currentLine+1
		fi
done

step=250
counter=0


for xpos in "-500" "500";
do
	for ypos in "-500" "500";
	do
		let nodeId=$counter+$numberOfCarriers

		echo	"	<node>
				<node_id>$nodeId</node_id>
				<position>
					<xpos>$xpos.000000</xpos>
					<ypos>$ypos.000000</ypos>
				</position>
			</node>" >> $output
		let counter=counter+1

	done
done


echo "	</node_settings>
	<mobility>" >> $output

timeStep="10"
let currentTime=$timeStep
countStatistics=0
rm -f *.dat

while [ "$currentTime" -lt "$simuTime" ];
do
	c=0
	while [ "$c" -lt "$numberOfCarriers" ];
	do
		delta=$(echo "${speed[$c]} * $timeStep" | bc)
		if [ "${xy[$c]}" = "x" ];
		then
			if [ "${versus[$c]}" = "n" ];
			then
				let xpos=${xpos[$c]}-$delta
				if [ "$xpos" -gt "-400" ];
				then
					xpos[$c]=$xpos
				else
					xpos[$c]="-400"
					xy[$c]="y"
					versus[$c]="p"
				fi
			else
				let xpos=${xpos[$c]}+$delta
				if [ "$xpos" -lt "400" ];
				then
					xpos[$c]=$xpos
				else
					xpos[$c]="400"
					xy[$c]="y"
					versus[$c]="n"
				fi
			fi	
			
		else
				
			if [ "${versus[$c]}" = "n" ];
			then
				let ypos=${ypos[$c]}-$delta
				if [ "$ypos" -gt "-400" ];
				then
					ypos[$c]=$ypos
				else
					ypos[$c]="-400"
					xy[$c]="x"
					versus[$c]="n"
				fi
			else
				let ypos=${ypos[$c]}+$delta
				if [ "$ypos" -lt "400" ];
				then
					ypos[$c]=$ypos
				else
					ypos[$c]="400"
					xy[$c]="x"
					versus[$c]="p"
				fi
			fi
		fi
		echo ${xpos[$c]} >> x-$c.dat
		echo ${ypos[$c]} >> y-$c.dat
		echo "<position_change>
				<node_id>$c</node_id>
				<start_time>$currentTime.000000</start_time>
				<destination>
					<xpos>${xpos[$c]}.000000</xpos>
					<ypos>${ypos[$c]}.000000</ypos>
				</destination>
				<velocity>${speed[$c]}.000000</velocity>
			</position_change>" >> $output
		let c=c+1
		let countStatistics=countStatistics+1
	done
	let currentTime=$currentTime+$timeStep
	
done  

echo "</mobility>
	<statistics>
		<stoptime>$simuTime</stoptime>
	</statistics>
</simulation>" >> $output
