#include <stdio.h>
#include <string.h>
#include <string>
#include <iostream>
#include <fstream>
#include "omnetpp.h"
#include "host.h"
#include "mobMessage_m.h"
#include <math.h> // ceil()

using namespace std;


Define_Module(Host);



bool Host::getIsInReach(unsigned int k)
{
	if (k>= (unsigned int)m_iNumHosts)
		throw cRuntimeError("Array of size %d indexed by %d",m_iNumHosts,k);

	return this->m_bIsInReach[k];
}


void Host::setIsInReach (unsigned int k, bool value)
{
	if (k>= (unsigned int) m_iNumHosts)
		throw cRuntimeError("Array of size %d indexed by %d",m_iNumHosts,k);

	this->m_bIsInReach[k]=value;
}


void Host::initialize()
{

	int i;

	m_parent=getParentModule();
	m_bRandomChoice=m_parent->par("randomChoice").boolValue();
	m_iNumHosts=m_parent->par("numberOfHosts");
	m_iInfoAndRoutingTableSize=m_parent->par("infoAndRoutingTableSize");
	m_iHistoryTablesSize=m_parent->par("historyTablesSize");
	m_iMsgBufSize=m_parent->par("msgBufSize");
	m_iMaxHops=m_parent->par("maxHops");
    m_iMaxTotalHops=m_parent->par("totalMaxHops");
	m_dRetransmissionInterval=m_parent->par("retransmissionInterval").doubleValue();
	m_dAutoRefreshTime=m_parent->par("autoRefreshTime").doubleValue();
	m_dRoutingTablesTransmissionInterval=m_parent->par("routingTablesTransmissionInterval").doubleValue();
	m_dColocW=m_parent->par("colocW").doubleValue();
	m_dCRCW=m_parent->par("CRCW").doubleValue();
	m_iProtocol = m_parent->par("protocol");
	//m_iLastRTTSize = m_parent->par("lastRTTSize");
	m_dProcTimeDelay=0.001;
	m_dRefreshTime=60.0;
	m_dCurrentRTT = 0.0;

	m_iInitialNumHops=m_iMaxHops;
    m_iInitialTotalNumHops=m_iMaxTotalHops;

	m_iMsgBufPointer=0;
	m_iInfoAndRoutingTablePointer=0;
	m_dCRCCorrelationCoefficient=0.0;
	m_iCRCHistoryPointer=-1;
	m_iCurrentHistoryTableSize=0;
	m_iNumMessagesReceived=0;
	m_iNumMessagesSent = 0;
	m_iNumMessagesToBeSent=0;
	m_iNumberOfStoredMessages = 0;
	m_iIterationNum=0;
	m_dPartialSumColoc=0.0;
	m_dNumSamplesSumColoc=0.0;
	m_dPartialSumCRC=0.0;
	m_dNumSamplesSumCRC=0.0;

	MAX_DISTANCE = 8;

	//m_ReputationVector.setName("Reputation Vector");
	m_TimerVector.setName("Timer Vector");
	m_RTTVector.setName("RTT Vector");
	m_DeliveryRatioVector.setName("Delivery Ratio");
	m_iCollectTimer = 50;
	m_dEpsilon=0.00;
	
	for ( i = 0; i < m_iNumHosts; i++)
	{
		m_iSelfUpdateCounter[i]=0;
		m_localPredictor.colocationDeliveryProb[i] = 0.0;
		m_localPredictor.deliveryProb[i] = 0.0;
	}

	for ( i = 0; i < m_iHistoryTablesSize; i++ )
		m_dCRCHistory[i]=0;
	
	m_dataCollector=(DataCollector*)(m_parent->getSubmodule("dataCollector"));
	m_iHostIdNumber=getIndex()+1;

	m_dReputationTableUpdateInterval = m_parent->par("reputationTableUpdateInterval");
	m_iNumberOfMessages = m_parent->par("numberOfMessages");
	m_iNumberOfCarriers = m_parent->par("numberOfCarriers");
	m_dInitialReputation = m_parent->par("initialReputation");
	m_dReputationIncrement = m_parent->par("reputationIncrement");
	m_dReputationDecrement = m_parent->par("reputationDecrement");
	m_iNumBlackholeCarriers = m_parent->par("numberOfActiveNonCollaborativeHosts");
	m_dBlackholeTreshold = m_parent->par("blackholeTreshold").doubleValue();
	//m_iRTTNumElements = 0;
	//m_dRTTPartialSum = 0;
	m_iReputationTableSize = m_iNumberOfCarriers;

	for( i = 0; i < m_iReputationTableSize; i++ )
	{
		std::string name("Reputation");
		name += i +'0';
		m_reputationTable[i].reputation = m_dInitialReputation;
		m_reputationTable[i].delta = 0;
		m_ReputationVector[i].setName(name.c_str());

	}

	m_iReachTableSize = m_iNumHosts;
	for( i = 0; i < m_iReachTableSize; i++ )
	{
		for(int j = 0; j < m_iReachTableSize; j++)
		{
			m_reachTable[i][j].deliveryProb = 2.0;
			m_reachTable[i][j].carrier = 0;
		}
	}
	
	for ( i = 0;i < m_iNumHosts; i++)
	{
		m_bWasInReach[i]=false;
		m_bIsInReach[i]=false;
	}

	m_bBufferFull=false;

	
	for ( i = 0; i < m_iMsgBufSize; i++)
	{
		m_msgBuf[i].messageId=0;
		m_msgBuf[i].recipient=0;
	}
	//cout<<"Buffer size: "<<m_iMsgBufSize<<"\n";
	//the info and routing table is inizialized with 0s
	for ( i = 0; i < m_iInfoAndRoutingTableSize; i++)
	{
		m_infoAndRoutingTable[i].neighborHostId=0;
		m_infoAndRoutingTable[i].targetHostId=0;
		m_infoAndRoutingTable[i].distance=0;
		m_infoAndRoutingTable[i].carrierHostId=0;
		m_infoAndRoutingTable[i].reputation = 0;
		m_infoAndRoutingTable[i].deliveryProb=0;
		m_infoAndRoutingTable[i].deliveryPred.currentPredictedValue=0;
		m_infoAndRoutingTable[i].refreshed=true;
	}

	//adding in position 0 (never deleted) the entry related to this host
	m_infoAndRoutingTable[0].neighborHostId=m_iHostIdNumber;
	m_infoAndRoutingTable[0].targetHostId=m_iHostIdNumber;
	m_infoAndRoutingTable[0].distance=0;
	m_infoAndRoutingTable[0].carrierHostId=m_iHostIdNumber;
	m_infoAndRoutingTable[0].reputation = 1.0;
	m_infoAndRoutingTable[0].deliveryProb=1;
	m_infoAndRoutingTable[0].deliveryPred.currentPredictedValue=1;
	m_infoAndRoutingTable[0].refreshed=true;

	for ( i=0; i < m_iHistoryTablesSize; i++ )
		m_dCRCHistory[i]=0;
	
	//for( i = 0; i < m_iLastRTTSize; i++ )
	//	m_dLastRTTVector[i] = 0.0;

	//m_iLastRTTPointer = 0;
	//m_iLastRTTNumElements = 0;

	m_RTTKalmanPred = new KalmanPredictor(m_dReputationTableUpdateInterval);
	mobMessage* autoMsg = new mobMessage("autoMsg");
	autoMsg->setType(AUTO_MSG);
	scheduleAt(simTime()+1.5, autoMsg);
	
	mobMessage* crcUpdateMsg = new mobMessage("crcUpdateMsg");
	crcUpdateMsg->setType(CRC_UPDATE_MSG);
	scheduleAt(simTime()+1.0,crcUpdateMsg);

	mobMessage* kalmanMsg = new mobMessage("kalmanMsg");
	kalmanMsg->setType(KALMAN_FILTER_AND_COLOCATION_UPDATE_MSG);
	scheduleAt(simTime()+0.5,kalmanMsg);

	mobMessage* infoAndRoutingMsg = new mobMessage("infoAndRoutingMsg");
	infoAndRoutingMsg->setType(INFO_AND_ROUTING_TABLE_EXCHANGE_MSG);
	scheduleAt(simTime()+2.5,infoAndRoutingMsg);

	mobMessage* bufferFlushMsg = new mobMessage("bufferFlushMsg");
	bufferFlushMsg->setType(LOCAL_BUFFER_FLUSH_MSG);
	scheduleAt(simTime()+initialWaitTime + 3.0,bufferFlushMsg);
	if(m_iProtocol != CAR){
		mobMessage* reputationTableUpdateMsg = new mobMessage("reputationTableUpdateMsg");
		reputationTableUpdateMsg->setType(TIMEOUT_REPUTATION_TABLE);
		scheduleAt(simTime() + initialWaitTime + 0.1 + m_dReputationTableUpdateInterval, reputationTableUpdateMsg);
	}

	mobMessage* collectMsg = new mobMessage("collectMsg");
	collectMsg->setType(COLLECT_MSG);
	scheduleAt(simTime() + m_iCollectTimer, collectMsg);
}


// AUTO_MSG
void Host::vHandleMessage_Type0(cMessage* msg)
{
	//first phase
	//updating the distance of the hosts that were in Reach and now are disconnected
	//distance=16 (RIP infinite)
	//and update context indicators

	mobMessage* msgReceived=static_cast<mobMessage*>(msg);

	for (int i=0;i<m_iNumHosts;i++)
	{
		if ((m_bIsInReach[i]==false))
		{
			for (int j=0;j<m_iInfoAndRoutingTableSize;j++)
			{
				if (i==(m_infoAndRoutingTable[j].neighborHostId-1))
				{
					m_infoAndRoutingTable[j].distance=MAX_DISTANCE;

				}//if
			}//for (int j=0
		}else
		{
			m_reputationTable[i].delta++;
		}
	//end if
	}//for (int i=0..

	scheduleAt(simTime()+1.0,msgReceived);
}

// PROTOCOL_MSG
void Host::vHandleMessage_Type1(cMessage* msg)
{
	mobMessage* msgReceived=static_cast<mobMessage*>(msg);

	int iSender=msgReceived->getSender();
	int i,j;


	// reset reachTable entries (only sender's row)
	for(i = 0; i < m_iReachTableSize; i++)
	{
		m_reachTable[iSender-1][i].carrier = 0;
		m_reachTable[iSender-1][i].deliveryProb = 0.0;
	}
	
	for (i=0;i<m_iInfoAndRoutingTableSize;i++)
	{
		bool bFound=false;
		if (msgReceived->getHostIdList(i)!=0)
		{
			int iHostId=msgReceived->getHostIdList(i);
			int iDistance=msgReceived->getDistanceList(i);
			double dDeliveryProb=msgReceived->getDeliveryProb(i);
			int iCarrierHostId=msgReceived->getCarrierHostId(i);

			// updating reachability table
			// we always update the entry even if the deliveryProb value is lower
			// in a way to have fresh values
			m_reachTable[iSender-1][iHostId-1].carrier = iCarrierHostId;
			m_reachTable[iSender-1][iHostId-1].deliveryProb = dDeliveryProb;

			//setting refreshed=false
			for (j=0;j<m_iInfoAndRoutingTableSize;j++)
			{
				m_infoAndRoutingTable[j].refreshed=false;
			}

			//if found update the entry to the local context info and routing table
			for (j=0;j<m_iInfoAndRoutingTableSize;j++)
			{
				//updating distances
				if ((m_infoAndRoutingTable[j].targetHostId==iHostId) && (m_infoAndRoutingTable[j].distance>iDistance))
				{
						m_infoAndRoutingTable[j].neighborHostId=iSender;
						m_infoAndRoutingTable[j].distance=iDistance+1;

				}//end if

				// calculate LUF (Local Utility Function) for the carrier
				// in received message and for the carrier in infoAndRoutingTable

				double dRep = m_reputationTable[iCarrierHostId-1].reputation;
				bool bUpdate = false;
				double dLUF1;
				double dLUF2;

				if(m_iProtocol != CAR){

					dLUF1 = dRep * dDeliveryProb;
					dLUF2 = m_reputationTable[(m_infoAndRoutingTable[j].carrierHostId)-1].reputation * m_infoAndRoutingTable[j].deliveryProb;
					if ((m_infoAndRoutingTable[j].targetHostId==iHostId)&&
					 ((dLUF1 - dLUF2) > 0))
					{
						bUpdate = true;

					}

				}else{


					if ( (m_infoAndRoutingTable[j].targetHostId==iHostId) &&
					 ((dDeliveryProb-m_infoAndRoutingTable[j].deliveryProb)>=0))
					{
						bUpdate = true;
					}
				}

				if(bUpdate){
					m_infoAndRoutingTable[j].deliveryProb=dDeliveryProb;
					m_infoAndRoutingTable[j].carrierHostId=iCarrierHostId;
					//cout<< "carrier host id: "<<iCarrierHostId<<"\n";
					m_infoAndRoutingTable[j].deliveryPred.reset();
					m_infoAndRoutingTable[j].deliveryPred.currentPredictedValue=m_infoAndRoutingTable[j].deliveryPred.getPredictedValue(dDeliveryProb);
					m_infoAndRoutingTable[j].refreshed=true;
					m_iSelfUpdateCounter[j]=0;
					m_infoAndRoutingTable[j].reputation = dRep;
					if (iDistance>MAX_DISTANCE)
						m_infoAndRoutingTable[j].distance=MAX_DISTANCE;
				}

				//updating probabilities (refresh of the information related
				//to the best host)
				if ((m_infoAndRoutingTable[j].carrierHostId==iCarrierHostId)&&
					(m_infoAndRoutingTable[j].targetHostId==iHostId))
				{
					if (iDistance < MAX_DISTANCE) {
						m_infoAndRoutingTable[j].deliveryProb=dDeliveryProb;
						m_infoAndRoutingTable[j].deliveryPred.currentPredictedValue=m_infoAndRoutingTable[j].deliveryPred.getPredictedValue(dDeliveryProb);
						m_infoAndRoutingTable[j].refreshed=true;
						m_iSelfUpdateCounter[j]=0;
						m_infoAndRoutingTable[j].reputation = dRep;
					}
				}

				if (m_infoAndRoutingTable[j].targetHostId==iHostId)
					bFound=true;


				/*[END ACK]*/

			}//end (int j=0;j<m_iInfoAndRoutingTableSize;j++ )

			bool bInserted=false;
			int p = 1;

			//if not found add to the list
			if ((bFound==false)&&(m_bBufferFull==true))
			{
				i = 0;
				while ((!bInserted)&&(i <m_iInfoAndRoutingTableSize))
				{
					if (m_infoAndRoutingTable[p].distance>MAX_DISTANCE)
						if (iDistance<MAX_DISTANCE)
						{
							m_infoAndRoutingTable[p].targetHostId=iHostId;
							m_infoAndRoutingTable[p].neighborHostId=iSender;
							m_infoAndRoutingTable[p].distance=iDistance+1;
							m_infoAndRoutingTable[p].carrierHostId=iCarrierHostId;
							m_infoAndRoutingTable[p].deliveryProb=dDeliveryProb;
							m_infoAndRoutingTable[p].reputation = m_reputationTable[iCarrierHostId-1].reputation;
							m_infoAndRoutingTable[p].deliveryPred.reset();
							m_infoAndRoutingTable[p].deliveryPred.currentPredictedValue=dDeliveryProb;
							m_infoAndRoutingTable[p].refreshed=true;

							bInserted=true;
						}

					if (p==m_iInfoAndRoutingTableSize-1)
						p=1;
					else
						p++;

					i++;
				}
			}

			if ((bFound==false)&&(bInserted==false)&&(m_bBufferFull==false))
			{
				if (m_iInfoAndRoutingTablePointer==m_iInfoAndRoutingTableSize-1)
				{
					m_iInfoAndRoutingTablePointer=1;
					m_bBufferFull=true;
				}
				//in position 0 we store the entry related to this host
				else
					m_iInfoAndRoutingTablePointer++;

				m_infoAndRoutingTable[m_iInfoAndRoutingTablePointer].targetHostId=iHostId;
				m_infoAndRoutingTable[m_iInfoAndRoutingTablePointer].neighborHostId=iSender;
				m_infoAndRoutingTable[m_iInfoAndRoutingTablePointer].distance=iDistance+1;
				m_infoAndRoutingTable[m_iInfoAndRoutingTablePointer].carrierHostId=iCarrierHostId;
				m_infoAndRoutingTable[m_iInfoAndRoutingTablePointer].deliveryProb=dDeliveryProb;
				m_infoAndRoutingTable[m_iInfoAndRoutingTablePointer].reputation = m_reputationTable[iCarrierHostId-1].reputation;
				m_infoAndRoutingTable[m_iInfoAndRoutingTablePointer].deliveryPred.reset();

				m_infoAndRoutingTable[m_iInfoAndRoutingTablePointer].deliveryPred.currentPredictedValue=dDeliveryProb;
				m_infoAndRoutingTable[m_iInfoAndRoutingTablePointer].refreshed=true;

			}
		}
	}


	delete msg;
}

bool Host::bSendSynchronous(cMessage* msg){
	mobMessage* msgReceived=static_cast<mobMessage*>(msg);
	int numHops=msgReceived->getNumHops();
	
	// if I must send an ack, the receiver is the original sender.
	int crossedHostsListPointer = msgReceived->getCrossedHostsListPointer();

	int recipient= msgReceived->getRecipient();

	// if the message has an intermediate recipient, it must be sent to it and not to the receiver
	int iIntermediateRecipient = msgReceived->getIntermediateRecipient();
	if( iIntermediateRecipient != 0 && m_iHostIdNumber != iIntermediateRecipient)
	{
		//cout<<"intermediate recipient is: "<<iIntermediateRecipient<<"\n";
		recipient = iIntermediateRecipient;
	//	msgReceived->setRecipient(iIntermediateRecipient);

	}

	for (int i=0;i<m_iInfoAndRoutingTableSize;i++)
	{
		if ((recipient==m_infoAndRoutingTable[i].targetHostId)&&
		(m_infoAndRoutingTable[i].distance < MAX_DISTANCE) &&  (m_infoAndRoutingTable[i].neighborHostId)!= m_iHostIdNumber &&
		(numHops>0) /*&&
		(totalNumHops>0)*/)
		{ 
			msgReceived->setNumHops(numHops - 1);
			bool alreadyInList = false;
			for(int j = 0; j < crossedHostsListPointer; j++)
			{
				//cout<<"neighbor: "<<m_infoAndRoutingTable[i].neighborHostId<<" receiver "<<recipient<<"\n";
 				if(msgReceived->getCrossedHostsList(j) == m_iHostIdNumber)
				{
 					alreadyInList = true;
					break;
				}
			}
			// add the node if is a carrier and is not already in list
			if(m_iHostIdNumber <= m_iNumberOfCarriers && !alreadyInList)
			{
				msgReceived->setCrossedHostsList(crossedHostsListPointer, m_iHostIdNumber);
				msgReceived->setCrossedHostsListPointer(crossedHostsListPointer+1);
			}

			if(m_iProtocol != CAR && m_reputationTable[m_infoAndRoutingTable[i].neighborHostId-1].reputation <= 0 && m_reachTable[m_infoAndRoutingTable[i].neighborHostId-1][recipient].deliveryProb > 0.9)
				return false;
			sendDelayed(msgReceived,m_dProcTimeDelay,"outputs",m_infoAndRoutingTable[i].neighborHostId-1);

			m_dataCollector->m_iTotalNumberOfSyncMessagesSent++;
			return true;
		}

	}
	return false;
}

bool Host::bSendAsynchronous(cMessage* msg){
	
	mobMessage* msgReceived=static_cast<mobMessage*>(msg);
	
	int recipient=msgReceived->getRecipient();
	int numHops=msgReceived->getNumHops();
//	int totalNumHops=msgReceived->getTotalNumHops();
	int crossedHostsListPointer = msgReceived->getCrossedHostsListPointer();
	

	//we search for the host is able to guarantee the best delivery probability
	for (int i=0;i<m_iInfoAndRoutingTableSize;i++)
	{
		if ((recipient==m_infoAndRoutingTable[i].targetHostId) && (m_infoAndRoutingTable[i].carrierHostId!=m_iHostIdNumber) /*&&
				((m_infoAndRoutingTable[i].deliveryProb-m_localPredictor.deliveryProb[i])>m_dEpsilon)*/
				&&
				(m_infoAndRoutingTable[i].carrierHostId!=0))
				//(numHops>0)
				/*&&*/


		{
			//cout<<"host id number: "<< m_iHostIdNumber << "carrier host id: "<< m_infoAndRoutingTable[i].carrierHostId<<" delivery prob: "<<
			//		m_infoAndRoutingTable[i].deliveryProb<<"\n";
					//m_dataCollector->m_iTotalNumberOfAsyncMessagesSent++;

			msgReceived->setNumHops(numHops - 1);
			bool alreadyInList = false;
			//int iHost;
			for(int j = 0; j < crossedHostsListPointer; j++)
			{

 				if(msgReceived->getCrossedHostsList(j) == m_iHostIdNumber)
				{
 					alreadyInList = true;
					break;
				}
			}
			// add the node if is a carrier and is not already in list
			if(m_iHostIdNumber <= m_iNumberOfCarriers && !alreadyInList)
			{
				msgReceived->setCrossedHostsList(crossedHostsListPointer, m_iHostIdNumber);
				msgReceived->setCrossedHostsListPointer(crossedHostsListPointer+1);
			}
			
			/** search how to reach the carrier **/
			for (int w=0;w<m_iInfoAndRoutingTableSize;w++)
			{
				if (m_infoAndRoutingTable[w].targetHostId==m_infoAndRoutingTable[i].carrierHostId)
				{
					msgReceived->setIntermediateRecipient(m_infoAndRoutingTable[i].carrierHostId);
					// the receiver contains the original receiver of the message, the recipient, the carrier.
					//msgReceived->setReceiver(recipient);
					//msgReceived->setRecipient(m_infoAndRoutingTable[i].carrierHostId);

					//cout<<"intermediate recipient: "<<m_infoAndRoutingTable[i].carrierHostId<<"\n";
					sendDelayed(msgReceived,m_dProcTimeDelay,"outputs",(m_infoAndRoutingTable[w].neighborHostId-1));
					m_dataCollector->m_iTotalNumberOfAsyncMessagesSent++;

					return true;
				}
			}//}
		} //if recipient
	}
	return false;
}

void Host::vStoreMessage(cMessage* msg){
	m_iNumberOfStoredMessages++;
	mobMessage* msgReceived=static_cast<mobMessage*>(msg);
	int recipient=msgReceived->getRecipient();
	int intermediateRecipient=msgReceived->getIntermediateRecipient();
	int messageId=msgReceived->getMessageId();
	int type = msgReceived->getType();
	int crossedHostsListPointer = msgReceived->getCrossedHostsListPointer();


	if(m_iMsgBufSize>0)
	{
		if (m_iMsgBufPointer==m_iMsgBufSize-1)
			m_iMsgBufPointer=0;
		else
		    	m_iMsgBufPointer++;
		m_iNumberOfStoredMessages++;
		m_msgBuf[m_iMsgBufPointer].protocolMessageType=1;
		m_msgBuf[m_iMsgBufPointer].messageId=messageId;
		m_msgBuf[m_iMsgBufPointer].recipient=recipient;
		m_msgBuf[m_iMsgBufPointer].intermediateRecipient= intermediateRecipient;
		m_msgBuf[m_iMsgBufPointer].sender=m_iHostIdNumber;
		m_msgBuf[m_iMsgBufPointer].timestamp=msgReceived->getTimestamp();
		m_msgBuf[m_iMsgBufPointer].numHops=m_iInitialNumHops;
		m_msgBuf[m_iMsgBufPointer].totalNumHops=m_iInitialTotalNumHops;
		m_msgBuf[m_iMsgBufPointer].type = type;
		for( int i = 0; i < crossedHostsListPointer; i++)
			m_msgBuf[m_iMsgBufPointer].crossedHostsList[i] = msgReceived->getCrossedHostsList(i);
		m_msgBuf[m_iMsgBufPointer].crossedHostsListPointer = crossedHostsListPointer;
		
	}
	delete msg;
}

// CONTROL_MSG
void Host::vHandleMessage_Type2(cMessage* msg)
{

	mobMessage* msgReceived=static_cast<mobMessage*>(msg);
	int recipient=msgReceived->getReceiver();
	msgReceived->setCrossedHostsListPointer(1);
	msgReceived->setCrossedHostsList(0,m_iHostIdNumber);
	msgReceived->setType(INFO_MSG);
	msgReceived->setRecipient(recipient);
	if ((m_iHostIdNumber==recipient)&&
		(recipient!=0))
	{
		return;
		
	}

	m_iNumMessagesSent++;

	msgReceived->setTimestamp(SIMTIME_DBL(simTime()));
	//if(!bSendSynchronous(msgReceived))
		if(!bSendAsynchronous(msgReceived))
		{
			vStoreMessage(msgReceived);
			
		}
}

// INFO_MSG
void Host::vHandleMessage_Type3(cMessage* msg)
{
	mobMessage* msgReceived=static_cast<mobMessage*>(msg);

	int recipient=msgReceived->getRecipient();
	//int recipient = msgReceived->getReceiver();
	int intermediateRecipient=msgReceived->getIntermediateRecipient();
	//cout<<"intermediate recipient: "<<intermediateRecipient<<" recipient: "<<recipient<<"\n";
	int messageId=msgReceived->getMessageId();
	int crossedHostsListPointer = msgReceived->getCrossedHostsListPointer();
	int nodeId;

	if(m_type == ACTIVE_NON_COLLABORATIVE)
	{
		//increase number of attracted messages excepted for acks
		if((m_iNumberOfMessages - messageId) >= 0 )
			m_dataCollector->m_iNumberOfAttractedMessages++;
		delete msg;
		return;
	}

	//update reputation
	int j = 0;
	double dDeliveryProb = 0.0;
	int iTarget = 0;
	double dRep = 0.0;

	if (m_iProtocol != CAR)
	{
		if(crossedHostsListPointer > 1){
			// increase reputation of the carriers
			for(int i = 1; i < crossedHostsListPointer; i++)
			{
				// get host id
				nodeId = msgReceived->getCrossedHostsList(i);

				m_reputationTable[nodeId-1].reputation = min(1.0, m_reputationTable[nodeId-1].reputation + m_dReputationIncrement);



				for(j = 0; j < m_iInfoAndRoutingTableSize; j++)
				{
					iTarget = m_infoAndRoutingTable[j].targetHostId;
					dDeliveryProb = m_reachTable[nodeId-1][iTarget-1].deliveryProb;
					dRep = m_reputationTable[nodeId - 1].reputation;
					if(dDeliveryProb*dRep > m_infoAndRoutingTable[j].deliveryProb*m_infoAndRoutingTable[j].reputation)
					{
						m_infoAndRoutingTable[j].deliveryProb = dDeliveryProb;
						m_infoAndRoutingTable[j].reputation = dRep;
						m_infoAndRoutingTable[j].carrierHostId = nodeId;
					}
				}

			}
		}
	}
	//data collector updating

	/** I AM THE RECEIVER **/
	if ((m_iHostIdNumber==recipient)&& (recipient!=0))
	{
		

		// if I am the receiver of an INFO_MSG build the ack.
		// if I am the receiver of an ACK_MSG update data collector

		// this is an INFO_MSG
		if((m_iNumberOfMessages - messageId) >= 0){

			double dDelay = SIMTIME_DBL(simTime()) - msgReceived->getTimestamp();
			m_dataCollector->vIncrementNumberOfMessagesReceived(msgReceived->getCrossedHostsList(0), dDelay);
			m_iNumMessagesReceived++;
			if(m_iProtocol != CAR){
				msgReceived->setRecipient(msgReceived->getCrossedHostsList(0));
				msgReceived->setIntermediateRecipient(0);
				msgReceived->setSender(m_iHostIdNumber);
				msgReceived->setNumHops(m_iInitialNumHops);
				msgReceived->setMessageId(messageId + m_iNumberOfMessages);
				//if(!bSendSynchronous(msgReceived)){
					if(!bSendAsynchronous(msgReceived))
						vStoreMessage(msgReceived);

			}
		}else{
			m_dataCollector->m_iNumberOfAckReceived++;
			m_dCurrentRTT = SIMTIME_DBL(simTime()) - msgReceived->getTimestamp();
			m_dReputationTableUpdateInterval = m_RTTKalmanPred->getPredictedValue(m_dCurrentRTT);

			delete msg;
		}
		
		return;
	}
	/** END I AM THE RECEIVER **/



	//this node is not the recipient (or the intermediate recipient of the message
	//in the case of the asynchronous routing), so we send it to another node
	//using the synchronous routing protocol or if it is not possible we look for the
	//host with the highest probability of delivering this message
	//Please note that the host with the highest delivery probability may be the node itself
	//(this node)


	/** I AM A WELL BEHAVING CARRIER **/
	if (m_iHostIdNumber==intermediateRecipient)
	{
		msgReceived->setIntermediateRecipient(0);
		if(m_type == ACTIVE_NON_COLLABORATIVE)
		{
			//increase number of attracted messages excepted for acks
			if((m_iNumberOfMessages - messageId) >= 0 )
				m_dataCollector->m_iNumberOfAttractedMessages++;
			delete msg;
			return;
		}
		else
		{
			m_dataCollector->m_iNumberOfForwardedMessages++;
			if(!bSendSynchronous(msg))
				if(!bSendAsynchronous(msg))
				{
					vStoreMessage(msg);
				}
			return;
		}
	}
	/** END I AM A WELL BEHAVING CARRIER **/

	if(!bSendSynchronous(msg)){
		//if(!bSendAsynchronous(msg)){
			vStoreMessage(msg);
		//}
	}
	
}

// CRC_UPDATE_MSG
void Host::vHandleMessage_Type4(cMessage* msg)
{
	mobMessage* msgReceived=static_cast<mobMessage*>(msg);

	//updating change degree of connectivitym_reputationTable[i].reputation
	double currentCRC;

	for (int w=0;w<m_iNumHosts;w++)
	{
		m_bWasInReach[w]=m_bIsInReach[w];
	}

	double cardUnion=0;
	for (int c=0;c<m_iNumHosts;c++)
	{
		if ((m_bWasInReach[c]==true)||(m_bIsInReach[c]==true))
		{
			if (c!=m_iHostIdNumber-1)
				cardUnion=cardUnion+1.0;
		}
	}

	double cardIntersection=0;

	for (int c=0;c<m_iNumHosts;c++)
	{
			if ((m_bWasInReach[c]==true)&&(m_bIsInReach[c]==true))
			{
				if (c!=m_iHostIdNumber-1)
					cardIntersection=cardIntersection+1.0;
			}
	}

	double totalIsInReach=0;
	for (int c=0;c<m_iNumHosts;c++)
	{
		if (c!=m_iHostIdNumber-1)
			totalIsInReach=totalIsInReach+1.0;
	}

	if (totalIsInReach==0)
		currentCRC=1.0;
		
	else
		currentCRC=(cardUnion-cardIntersection)/cardUnion;

	//updating CRC history pointer
	if (m_iCRCHistoryPointer<m_iHistoryTablesSize-1)
		m_iCRCHistoryPointer++;
	else
		m_iCRCHistoryPointer=0;

	m_dCRCHistory[m_iCRCHistoryPointer]=currentCRC;

	if (m_iCurrentHistoryTableSize<m_iHistoryTablesSize)
		m_iCurrentHistoryTableSize++;

	//calculation of the covariance
	int locPointer=m_iCRCHistoryPointer;
	double numberOfIteration=0.0;
	double partialSum=0.0;
	double partialSum_previous=0.0;
	double average;
	double average_previous;
	double CRCCovariance=0.0;
	double partialCovSum=0.0;
	double partialVarSum=0.0;
	double CRCVariance=0.0;

	while (numberOfIteration<=m_iCurrentHistoryTableSize)
	{
		numberOfIteration=numberOfIteration+1.0;

		if (locPointer>0)
			partialSum_previous=partialSum_previous+m_dCRCHistory[(locPointer-1)];
		else
			partialSum_previous=partialSum_previous+m_dCRCHistory[(m_iHistoryTablesSize-1)];

		partialSum=partialSum_previous+m_dCRCHistory[(locPointer)];

		if (numberOfIteration>2)
		{
			average_previous=partialSum_previous/(numberOfIteration);
			average=partialSum/(numberOfIteration+1.0);
		}
		else
		{
			average_previous=0.0;
			average=0.0;
		}


		if (locPointer>0)
			partialCovSum=partialCovSum+(m_dCRCHistory[(locPointer-1)]-average_previous)*(m_dCRCHistory[(locPointer)]-average);
		else
			partialCovSum=partialCovSum+(m_dCRCHistory[m_iHistoryTablesSize-1]-average_previous)*(m_dCRCHistory[(locPointer)]-average);

		partialVarSum=partialVarSum+(m_dCRCHistory[locPointer]-average)*(m_dCRCHistory[locPointer]-average);

		if (locPointer>0)
			locPointer--;
		else
			locPointer=m_iHistoryTablesSize-1;

	}

	CRCCovariance=partialCovSum/(numberOfIteration-1.0);
	CRCVariance=partialVarSum/(numberOfIteration);

	if ((numberOfIteration>3)&&(CRCVariance!=0))
		m_dCRCCorrelationCoefficient=(CRCCovariance)/(CRCVariance);
	else
		m_dCRCCorrelationCoefficient=0.0;//case numberOfIteration=0

	//change of the sign
	if (m_dCRCCorrelationCoefficient<0)
		m_dCRCCorrelationCoefficient=0.0-m_dCRCCorrelationCoefficient;


	m_localPredictor.CRCDeliveryProb=m_localPredictor.CRCPred.getPredictedValue(currentCRC);

	if (m_localPredictor.CRCPred.isPredictable())
		m_dPartialSumCRC=m_dPartialSumCRC+1.0;
	m_dNumSamplesSumCRC=m_dNumSamplesSumCRC+1.0;

	msgReceived->setType(CRC_UPDATE_MSG);
	scheduleAt(simTime()+30.0, msgReceived);
}

// KALMAN_FILTER_AND_COLOCATION_UPDATE_MSG
void Host::vHandleMessage_Type5(cMessage* msg)
{
	mobMessage* msgReceived=static_cast<mobMessage*>(msg);

	//prediction of the values using Kalman filter for the values that has not been
	//refreshed

	for (int i=0;i<m_iInfoAndRoutingTableSize;i++)
	{
		if ((m_infoAndRoutingTable[i].refreshed==false)&&
			(m_infoAndRoutingTable[i].targetHostId!=0))
		{
			m_infoAndRoutingTable[i].deliveryPred.currentPredictedValue=m_infoAndRoutingTable[i].deliveryPred.getPredictedValue(m_infoAndRoutingTable[i].deliveryPred.currentPredictedValue);
			//If not refreshed recently we use the deliveryProb that we have predicted

			m_infoAndRoutingTable[i].deliveryProb=m_infoAndRoutingTable[i].deliveryPred.currentPredictedValue;
			m_iSelfUpdateCounter[i]=m_iSelfUpdateCounter[i]++;

			/*if (m_iSelfUpdateCounter[i]==1)
			{
				m_infoAndRoutingTable[i].carrierHostId=0;
				m_infoAndRoutingTable[i].deliveryPred.currentPredictedValue=0;
			}*/
		}
	}

	//updating colocation
	//colocation value
	//0 not in Reach
	//1 in Reach
	double colocationValue;

	double localSumColoc=0.0;

	for (int c1=0;c1<m_iNumHosts;c1++)
	{
		bool inTheCloud=false;
		for (int c2=0;c2<m_iInfoAndRoutingTableSize;c2++)
		{
			if ((m_infoAndRoutingTable[c2].targetHostId==(c1+1))&&
				  (m_infoAndRoutingTable[c2].distance<MAX_DISTANCE))
				inTheCloud=true;
		}

		if (inTheCloud==true)
			colocationValue=1.0;
		else
			colocationValue=0.0;

		m_localPredictor.colocationDeliveryProb[c1]=m_localPredictor.colocationPred[c1].getPredictedValue(colocationValue);

		if (m_localPredictor.colocationPred[c1].isPredictable()==true)
			localSumColoc=localSumColoc+1.0;

	}//for (int c1=0;...

	m_dPartialSumColoc=m_dPartialSumColoc+localSumColoc/(double)m_iNumHosts;
	m_dNumSamplesSumColoc=m_dNumSamplesSumColoc+1.0;

	//calculation of the total delivery probability
	for (int c=0;c<m_iNumHosts;c++)
	{
		m_localPredictor.deliveryProb[c]=m_dColocW*m_localPredictor.colocationDeliveryProb[c]+m_dCRCW*m_localPredictor.CRCDeliveryProb;
	}

	//update current probability
	for (int z=0;z<m_iInfoAndRoutingTableSize;z++)
	{
		if ((m_infoAndRoutingTable[z].targetHostId > 0)&&
				(m_localPredictor.deliveryProb[(m_infoAndRoutingTable[z].targetHostId)-1]>m_infoAndRoutingTable[z].deliveryProb))
		{
			m_infoAndRoutingTable[z].carrierHostId=m_iHostIdNumber;
			m_infoAndRoutingTable[z].deliveryProb=m_localPredictor.deliveryProb[(m_infoAndRoutingTable[z].targetHostId)-1];
			m_infoAndRoutingTable[z].refreshed=true;
			m_iSelfUpdateCounter[z]=0;
		}

		if (m_infoAndRoutingTable[z].carrierHostId==m_iHostIdNumber)
		{
			m_infoAndRoutingTable[z].deliveryProb=m_localPredictor.deliveryProb[(m_infoAndRoutingTable[z].targetHostId)-1];
			m_infoAndRoutingTable[z].refreshed=true;
			m_iSelfUpdateCounter[z]=0;
		}
	}//for

	//mobMessage *msg=new mobMessage("autoMessage");
	msgReceived->setType(KALMAN_FILTER_AND_COLOCATION_UPDATE_MSG);
	scheduleAt(simTime()+m_dAutoRefreshTime, msgReceived);
}

// INFO_AND_ROUTING_TABLE_EXCHANGE_MSG
void Host::vHandleMessage_Type6(cMessage* msg)
{
	mobMessage* msgReceived=static_cast<mobMessage*>(msg);

	//second phase: sending context and routing tables to the neighbours
	for (int k=0;k<m_iNumHosts;k++)
	{
		if (getIsInReach(k))
		{
			char messageName[26];

			
			sprintf(messageName,"%s%d","context&routingInfoFrom",m_iHostIdNumber);
			mobMessage* mymsg = new mobMessage(messageName);
			mymsg->setType(PROTOCOL_MSG);
			mymsg->setSender(m_iHostIdNumber);
			mymsg->setReceiver(k);
			mymsg->setNumHops(m_iInitialNumHops);
			
						
			int counter=0;
			int i;

			/*if(m_type == ACTIVE_NON_COLLABORATIVE)
			{	
				//for ( i = 0; i < m_iNumHosts; i++ )
				for ( i=0;i<m_iInfoAndRoutingTableSize;i++)
				{				
					if (m_infoAndRoutingTable[i].targetHostId!=0)
					{
					//mymsg->setDistanceList(i, 1);
					mymsg->setDistanceList(i,m_infoAndRoutingTable[i].distance);
					mymsg->setDeliveryProb(i, 1);
					mymsg->setCarrierHostId(i,m_iHostIdNumber);
					mymsg->setHostIdList(i,i);
					}
				}
			}
			else
			{*/

				for ( i=0;i<m_iInfoAndRoutingTableSize;i++)
				{

					if (m_infoAndRoutingTable[i].targetHostId!=0)
					{
						mymsg->setHostIdList(counter,m_infoAndRoutingTable[i].targetHostId);

						if(m_type == ACTIVE_NON_COLLABORATIVE)
						{
							mymsg->setDeliveryProb(counter, 1);
							mymsg->setCarrierHostId(counter, m_iHostIdNumber);
						}
						else
						{
							mymsg->setDeliveryProb(counter, m_infoAndRoutingTable[i].deliveryProb);
							mymsg->setCarrierHostId(counter,m_infoAndRoutingTable[i].carrierHostId);
						}
						mymsg->setDistanceList(counter,m_infoAndRoutingTable[i].distance);

						counter++;
					}
				}//for (int i=0;i<m_iInfoAndRoutingTableSize;i++)
			//}
			sendDelayed(mymsg,m_dProcTimeDelay,"outputs",k);
			m_dataCollector->m_iTotalNumberOfRoutingMessagesSent++;
		}//if (isInReach[k]... /

	}//for (int k=0...
	
	m_iIterationNum++;

	double variableRefreshTime;

	if (m_dCRCCorrelationCoefficient==0)
	{
		variableRefreshTime=m_dAutoRefreshTime*0.1;
	}
	else if (m_dCRCCorrelationCoefficient<0.1) 
		variableRefreshTime=m_dAutoRefreshTime*0.1;
	else
		variableRefreshTime=m_dAutoRefreshTime*m_dCRCCorrelationCoefficient;


	scheduleAt(simTime()+m_dRoutingTablesTransmissionInterval, msgReceived);

}

// LOCAL_BUFFER_FLUSH_MSG
void Host::vHandleMessage_Type7(cMessage* msg)
{
	mobMessage* msgReceived=static_cast<mobMessage*>(msg);

	//sending stored messages synch (if it is possible) or asynch (if there is
	//a host with a better delivery probability around.

	int recipient, carrier;

	for (int z=0;z<m_iMsgBufSize;z++)
	{
		recipient=m_msgBuf[z].recipient;
		carrier = m_msgBuf[z].intermediateRecipient;

		

		if (recipient!=0)
		{

			mobMessage* msgInfoNew = new mobMessage("infoMessage");
			msgInfoNew->setProtocolMessageType(1);
			msgInfoNew->setMessageId(m_msgBuf[z].messageId);
			msgInfoNew->setRecipient(m_msgBuf[z].recipient);
			msgInfoNew->setIntermediateRecipient(m_msgBuf[z].intermediateRecipient);
			//cout<<"buffer flush intermediate recipient: "<<m_msgBuf[z].intermediateRecipient<<"\n";
			msgInfoNew->setTimestamp(m_msgBuf[z].timestamp);
			msgInfoNew->setSender(m_iHostIdNumber);
			msgInfoNew->setType(m_msgBuf[z].type);
			msgInfoNew->setNumHops(m_iInitialNumHops);
			int totalNumHops=m_msgBuf[z].totalNumHops;
			msgInfoNew->setTotalNumHops(totalNumHops-1);

			int crossedHostsListPointer = m_msgBuf[z].crossedHostsListPointer;
			bool alreadyInList = false;

			for(int j = 0; j < crossedHostsListPointer; j++)
			{
				msgInfoNew->setCrossedHostsList(j, m_msgBuf[z].crossedHostsList[j]);
					}


				if(!bSendSynchronous(msgInfoNew))
					//if(!bSendAsynchronous(msgInfoNew))
					{
						delete msgInfoNew;
						continue;
					}


			//removing the message from the message_buffer
			//currently not a dynamic list
			m_msgBuf[z].recipient=0;
			m_msgBuf[z].sender=0;
			m_msgBuf[z].messageId=0;
			m_msgBuf[z].intermediateRecipient=0;
			m_msgBuf[z].numHops=0;
			m_msgBuf[z].totalNumHops=0;
			m_msgBuf[z].protocolMessageType=0;
			m_msgBuf[z].timestamp=0;

			for(int j = 0; j < crossedHostsListPointer; j++)
			{
				m_msgBuf[z].crossedHostsList[j] = 0;
			}
			m_msgBuf[z].crossedHostsListPointer = 0;
		}
	}
	scheduleAt(simTime()+m_dRetransmissionInterval, msgReceived);
}

// TIMEOUT_REPUTATION_TABLE
void Host::vHandleMessage_Type9(cMessage* msg)
{
	/* update reputation table entries */
	for(int i = 0; i < m_iReputationTableSize; i++)
	{
		if(i == (m_iHostIdNumber-1))
			continue;

		//if(m_reputationTable[i].delta > 0)
			m_reputationTable[i].reputation = max(0.0, m_reputationTable[i].reputation - m_dReputationDecrement);

	}

	for (int i = 0; i < m_iInfoAndRoutingTableSize; i++)
	{
		m_infoAndRoutingTable[i].reputation = m_reputationTable[m_infoAndRoutingTable[i].carrierHostId-1].reputation;
	}
	scheduleAt(simTime() + m_dReputationTableUpdateInterval, msg);
}

// ROUTING_TABLE_DBG
void Host::vHandleMessage_Type10(cMessage* msg)
{
	double tmp = SIMTIME_DBL(simTime());

	cout<<"Routing table of the host "<<m_iHostIdNumber<<" at time "<<tmp<<"\n";
	cout<<"Target     neighbor   distance      carrierHostId    deliveryProb  predDeliveryProb"<<"\n";

	for (int r=0;r<m_iInfoAndRoutingTableSize;r++)
		cout<<m_infoAndRoutingTable[r].targetHostId<<"             "<<m_infoAndRoutingTable[r].neighborHostId<<"             "<<m_infoAndRoutingTable[r].distance<<"              "<<m_infoAndRoutingTable[r].carrierHostId<<"              "<<m_infoAndRoutingTable[r].deliveryProb<<"            "<<m_infoAndRoutingTable[r].deliveryPred.currentPredictedValue<<"\n";
	cout<<"\n";

	scheduleAt(simTime() + 1000, msg);
}


void Host::handleMessage(cMessage *msg)
{
	// Message type ("type" field of mobMessage)
	// code 0 autoMessage (used for observing if the "context" has changed --periodically--)
	// code 1 protocolMessage
	// code 2 controlMessage
	// code 3 informationMessage
	// code 4 CRC update (periodically)
	// code 5 update Kalman filters
	// code 6 send routing information table
	// code 7 send messages stored in the local buffer

	mobMessage* msgReceived = static_cast<mobMessage*>(msg);

	// retrieving message type
	int type = msgReceived->getType();

	switch(type)
	{
		case AUTO_MSG:
			vHandleMessage_Type0(msg);
			break;
		case PROTOCOL_MSG:
			vHandleMessage_Type1(msg);
			break;
		case CONTROL_MSG:
			vHandleMessage_Type2(msg);
			break;
		case INFO_MSG:
			ev<<"INFO\n";
			vHandleMessage_Type3(msg);
			ev<<"END INFO\n";
			break;
		case CRC_UPDATE_MSG:
			ev<<"CRC\n";
			vHandleMessage_Type4(msg);
			ev<<"CRC\n";
			break;
		case KALMAN_FILTER_AND_COLOCATION_UPDATE_MSG:
			ev<<"KALMAN\n";
			vHandleMessage_Type5(msg);
			ev<<"END KALMAN\n";
			break;
		case INFO_AND_ROUTING_TABLE_EXCHANGE_MSG:
			vHandleMessage_Type6(msg);
			break;
		case LOCAL_BUFFER_FLUSH_MSG:
			ev<<"BUFFER\n";
			vHandleMessage_Type7(msg);
			ev<<"BUFFER\n";
			break;
		case TIMEOUT_REPUTATION_TABLE:
			vHandleMessage_Type9(msg);
			break;
		case ROUTING_TABLE_DBG:
			vHandleMessage_Type10(msg);
			break;
		case COLLECT_MSG:
			int i = 0;
			for(i = 0; i < m_iNumHosts; i++)
				m_ReputationVector[i].record(m_reputationTable[i].reputation);
			m_TimerVector.record(m_dReputationTableUpdateInterval);
			m_RTTVector.record(m_dCurrentRTT);
			if(m_iNumMessagesSent > 0)
				m_DeliveryRatioVector.record(double(m_iNumMessagesReceived)/double(m_iNumMessagesSent));
			else
				m_DeliveryRatioVector.record(0);
			scheduleAt(simTime() + m_iCollectTimer, msg);
			break;
	}

} // end handleMessage


void Host::finish()
{

	int iNumberOfBlackhole = 0;
	int iNumberOfFalsePositive = 0;
	int i;

	for ( i = 0; i < m_iNumberOfCarriers; i++ )
	{
		recordScalar("reputation", m_reputationTable[i].reputation);
		recordScalar("delta", m_reputationTable[i].delta);
		if ( m_reputationTable[i].reputation < m_dBlackholeTreshold &&
		     m_reputationTable[i].delta > 0 )
		{
			if ( i < m_iNumBlackholeCarriers )
				iNumberOfBlackhole++;
			else
				iNumberOfFalsePositive++;

		}
	}

	recordScalar("Number of Detected Blackhole", iNumberOfBlackhole);
	recordScalar("Number of False Positive", iNumberOfFalsePositive);
	recordScalar("Number of Stored Messages", m_iNumberOfStoredMessages);
	recordScalar("Average reputation update interval",m_dReputationTableUpdateInterval);
}

