
#include "kalmanPredictor.h"
#include "dataCollector.h"
#include "definitions.h"
#include <coutvector.h>

class Host: public cSimpleModule {

	struct msgBufItem{
		int recipient;
		int intermediateRecipient;
		int sender;
		int messageId;
		int numHops;
		int totalNumHops;
		int protocolMessageType;
		double timestamp;
		char name[20];
		int type;
		int crossedHostsList[500];
		int crossedHostsListPointer;
	};

	msgBufItem m_msgBuf[BUFFER_SIZE];

	//max number of hops during a synchnous transmission
	int m_iMaxHops;

	//max number of retransmissions
	int m_iMaxTotalHops;

	bool m_bIsInReach[MAX_NUM_HOSTS];
	//delay due to processing time
	double m_dProcTimeDelay;

	//random choice
	bool m_bRandomChoice;

	//buffer is full
	bool m_bBufferFull;

	//maximum refresh time of the message tables
	double m_dRefreshTime;

	//refresh time of the routing table (if necessary, using probabilistic method)
	double m_dAutoRefreshTime;

	//number of Hosts
	int m_iNumHosts;

	//update counter
	int m_iSelfUpdateCounter[MAX_NUM_HOSTS];

	//pointer to module
	cModule* m_parent;

	//infoAndRouting Table Size
	int m_iInfoAndRoutingTableSize;

	//history table size
	int m_iHistoryTablesSize;

	//message buffer size
	int m_iMsgBufSize;

	//retransmissionInterval
	double m_dRetransmissionInterval;

	//auto refresh time

	//routing table transmission interval
	double m_dRoutingTablesTransmissionInterval;

	//epsilon for asynchronous send
	double m_dEpsilon;

	double m_dColocW;

	double m_dCRCW;

	int m_iProtocol;

	cOutVector m_ReputationVector[MAX_NUM_HOSTS];
	cOutVector m_TimerVector;
	cOutVector m_RTTVector;
	int m_iCollectTimer;
public:

	nodeType m_type;

	int m_iNumBlackholeCarriers;

	// number of carriers
	int m_iNumberOfCarriers;


	double m_dInitialReputation;
	double m_dReputationIncrement;
	double m_dReputationDecrement;
	// initial group the host belongs to: for carriers it will change over time
	int m_dStartingGroup;

	bool m_bWasInReach[MAX_NUM_HOSTS];

	int MAX_DISTANCE;

	//item of the info and routing table
	//used to store information about the other hosts
	struct infoAndRoutingTableItem {
		int targetHostId;
		int neighborHostId;
		int distance;//distance from targetHostId passing through targetHostId
   	 	int carrierHostId;
		double deliveryProb;
		double reputation; // reputation of the carrier
		KalmanPredictor deliveryPred;
		bool refreshed;
	};

	//item of the table used to store information/prediction about the colocation with other hosts
	struct localPredictorItem {
		double CRCDeliveryProb;
		KalmanPredictor CRCPred;

		double colocationDeliveryProb[MAX_NUM_HOSTS];
		KalmanPredictor colocationPred[MAX_NUM_HOSTS];

		double deliveryProb[MAX_NUM_HOSTS];
	};

	localPredictorItem m_localPredictor;

	//initial number of hops
	int m_iInitialNumHops;

	int m_iInitialTotalNumHops;

	//pointer of the buffer of the messages
	int m_iMsgBufPointer;

	//pointer of the buffer of the context info and routing table
	int m_iInfoAndRoutingTablePointer;

	//correlation coefficient
	double m_dCRCCorrelationCoefficient;

	//pointer of the buffer of the CRC history
	int m_iCRCHistoryPointer;

	//number of valid values in the current history table size
	//the maximum number is historyTableSize (see constants)
	int m_iCurrentHistoryTableSize;

	//history (previous samples) of CRC
	double m_dCRCHistory[MAX_HISTORY_TABLE_SIZE];

	//id of the node
	int m_iHostIdNumber;

	//iteration num
	//number of iteration of the phase of sending mesages/routing tables
	int m_iIterationNum;

	//ref to dataCollector
	DataCollector* m_dataCollector;

	int m_iNumMessagesToBeSent;
	int m_iNumMessagesReceived;
	int m_iNumMessagesSent;
	cOutVector m_DeliveryRatioVector;

	int m_iNumberOfStoredMessages;

	infoAndRoutingTableItem m_infoAndRoutingTable[MAX_NUM_HOSTS];

	//variables used for calculating the predictability of the time series of colocation
	double m_dPartialSumColoc;
	double m_dNumSamplesSumColoc;

	//variables used for calculating the predictability of the time series of the
	//change degree of connectivity
	double m_dPartialSumCRC;
	double m_dNumSamplesSumCRC;

	// total number of messages (.ini file parameter)
	int m_iNumberOfMessages;

	/* Reputation Table: stores reputation for each host.
	*/
	struct reputationTableItem
	{
		// store reputation as a double in [0,1]
		double reputation;
		// specify how many times this host has met the carrier
		int delta;
	
	};
	reputationTableItem m_reputationTable[MAX_NUM_HOSTS];
	int m_iReputationTableSize;
	double m_dBlackholeTreshold;

	// update interval used for the periodical decrement of reputation table's entries
	double m_dReputationTableUpdateInterval;
	double m_dCurrentRTT;
	KalmanPredictor* m_RTTKalmanPred;
	//double m_dLastRTTVector[MAX_RTT_NUMBER];
	//int m_iLastRTTSize;
	//int m_iLastRTTPointer;
	//int m_iLastRTTNumElements;

	//int m_iRTTNumElements;
	//double m_dRTTPartialSum;
	

	/* Reachability table:
	is a numCarriers x numHosts table, in which each host stores information about reachability
	of all the hosts in the network, collecting information from the PROTOCOL_MSGs received.
	Rows indicate the carriers, columns indicate each host in the network.

	reachTable[i][j] tells that carrier "i" can reach host "j"
	with a certain probability deliveryProb
	*/
	struct reachTableItem
	{
		double deliveryProb;
		int carrier;
	};
	reachTableItem m_reachTable[MAX_NUM_HOSTS][MAX_NUM_HOSTS];
	int m_iReachTableSize;

	virtual bool getIsInReach(unsigned int k);

	virtual void setIsInReach(unsigned int k, bool isInReach);

	virtual void initialize();

	virtual void handleMessage(cMessage* msg);

	virtual void finish();

	virtual void vHandleMessage_Type0(cMessage *msg);
	virtual void vHandleMessage_Type1(cMessage *msg);
	virtual void vHandleMessage_Type2(cMessage *msg);
	virtual void vHandleMessage_Type3(cMessage *msg);
	virtual void vHandleMessage_Type4(cMessage *msg);
	virtual void vHandleMessage_Type5(cMessage *msg);
	virtual void vHandleMessage_Type6(cMessage *msg);
	virtual void vHandleMessage_Type7(cMessage *msg);
	virtual void vHandleMessage_Type9(cMessage *msg);
	virtual void vHandleMessage_Type10(cMessage *msg);
	virtual bool bSendSynchronous(cMessage* msg);
	virtual bool bSendAsynchronous(cMessage* msg);
	virtual void vStoreMessage(cMessage* msg);
	
};

