#include <stdio.h>
#include <string.h>
#include <math.h>
#include <iostream>
#include "omnetpp.h"
#include "kalmanPredictor.h"



//Function tha returns the absolute value of a double
double abs(double value) {
	
	if (value>0) 
		return value;
	else
		return -value;	
	
	
}

void KalmanPredictor::init()
{
	//initialisation of the variables
	xPred1=0.5;
	qt=0.01;
	rt=0.1;
	counter=1;
	
	//autocorrelation variables
	totalnumberOfSamples=0;
	
	//number of residuals used in the calculation of the autocorrelation coefficients
	//numberOfSamples=55;
	
	//upper bound of the lag used in the calculation of the autocorrelation coefficients
	//k_max=30;
	
	//pointer to the current position of the residuals array
	pointer=0;
	
	//maximum acceptable Average Percentage Forecasting Error.
	APFE=0.20;
	
	//private enable logging
	logging=false;
	
	//samples
	//samples=new double[numberOfSamples];
	
	//residuals used in the calculation of the autocorrelation coefficients
	//residuals= new double[numberOfSamples];
	
	int i;
	for ( i = 0; i < NUMBER_OF_SAMPLES; i++)
	{
		residuals[i] = 0.0;
		samples[i] = 0.0;
		pError[i] = 0.0;

	}

	for ( i = 0; i < K_MAX; i++ )
		r[i] = 0;

	//average percentage error used in the calculation of the forecasting error 
	//pError=new double[numberOfSamples];
	
	//autocorrelation coefficients
	//r=new double[(int)k_max];
	
	numberOfValidSamples=0;	
	numberOfPredictableSamples=0;
}

KalmanPredictor::KalmanPredictor(double dInitialPredictedValue)
{

	INITIAL_PREDICTED_VALUE = dInitialPredictedValue;
	init();
	//qt=100;
}

KalmanPredictor::KalmanPredictor()
{

	INITIAL_PREDICTED_VALUE = 0.5;
	init();

}

void KalmanPredictor::reset() {
	//initialisation of the variables
	xPred1=0.5;

	qt=0.01;
	rt=0.1;
	counter=1;
	
	//autocorrelation variables
	totalnumberOfSamples=0;
	
	//number of residuals used in the calculation of the autocorrelation coefficients
	//numberOfSamples=55;
	
	//upper bound of the lag used in the calculation of the autocorrelation coefficients
	//k_max=30;
	
	//pointer to the current position of the residuals array
	pointer=0;
	
	//maximum acceptable Average Percentage Forecasting Error.
	APFE=0.00;
	
}

double KalmanPredictor::getPredictedValue(double currentValue) {
	
		currentInputValue=currentValue;
		
		pastPredictedValue=currentPredictedValue;
		

		
	if (counter==1) {
		this->currentValue=currentValue;
		//currentPredictedValue=0.5;
		currentPredictedValue= INITIAL_PREDICTED_VALUE;
		//currentOmegat=((1.0-currentPredictedValue)*(1.0-currentPredictedValue)*(1.0-currentPredictedValue)-currentPredictedValue*currentPredictedValue*currentPredictedValue)/3.0;
		currentOmegat = 1;
		counter++;
	} else {
		
		this->currentValue=currentValue;
		currentPredictedValue=currentPredictedValue+currentOmegat*(currentValue-currentPredictedValue);///(currentOmegat+rt);
		currentOmegat=currentOmegat+qt-currentOmegat*currentOmegat/(currentOmegat+rt);
		counter++;
	}
	
	double residual=pastPredictedValue-currentInputValue;
	
	if ((totalnumberOfSamples<=NUMBER_OF_SAMPLES)){
		pointer=pointer%NUMBER_OF_SAMPLES;
		residuals[pointer]=residual;
		samples[pointer]=currentInputValue;
		pointer++;
		//total number of residuals in input
		totalnumberOfSamples++;
	}
	
	return currentPredictedValue;
}

/**
* This method returns true if the timeseries is predictable, false otherwise.
*/
bool KalmanPredictor::isPredictable() {

	double average;
	double x_t;
	double x_t_plus_k;
	int p_plus_k;
		
	bool isPredictable=true;

		
	//the time series is not predictable if the number of samples is not sufficient...
	if (totalnumberOfSamples<NUMBER_OF_SAMPLES) {
		isPredictable=false;
		return isPredictable;
	} 
	else {
		numberOfValidSamples=numberOfValidSamples+1.0;
	}
		
	int i=0;
	int j=0;
	double tot=0;
	double num=0;
	double den=0;
	
	//calculation of the average
	for (i=0;i<NUMBER_OF_SAMPLES;i++) {
		tot=tot+residuals[i];
				}
	average=tot/(double)(NUMBER_OF_SAMPLES);
		
		
		
     //*****************************************************
	//Analysis of the average percentage forecasting error
	//*****************************************************
		
	//average percentage error
	for (i=0;i<NUMBER_OF_SAMPLES;i++) {
		if (residuals[i]==0)
			pError[i]=0;
		else {
			pError[i]=residuals[i];
			//if (samples[i]!=0)
			//	pError[i]=abs(residuals[i])/abs(samples[i]);
			//else
			//	pError[i]=1.0;
				//pError[i]=abs(residuals[i])/abs(samples[i]);
		}
	}
	double totPError=0.0;
	
	double numberOfSamples2=NUMBER_OF_SAMPLES;
	for (i=0;i<NUMBER_OF_SAMPLES;i++)
		if (abs(pError[i])>0.4) 
			numberOfSamples2--;
	    else
		totPError=totPError+abs(pError[i]);
	double averagePError=totPError/(double)(numberOfSamples2);
		
	
		
	if (averagePError>0.15) 
		isPredictable=false;

		
	//************************************
	//Residual Analysis - Autocorrelation
	//************************************
		
	//calculation of the denominator of r_k
	//the denominator is the same for every r_k
	for (j=0;j<NUMBER_OF_SAMPLES;j++){
		den=den+(residuals[j]-average)*(residuals[j]-average);
		//System.out.println(residuals[j]);
	}
		
	//calculation of the numerator of r_k for k=0...k_max
	int k=0;
	for (k=0;k<K_MAX;k++) {
		num=0;
		//p=pointer;
		for (j=0;j<NUMBER_OF_SAMPLES-k;j++) {
			//x_t=residuals[p];
			//p_plus_k=(p+k)%numberOfSamples;
			//x_t_plus_k=residuals[p_plus_k];
			//num=num+(x_t-average)*(x_t_plus_k-average);
			//p++;
			//p=p%numberOfSamples;
			
			x_t=residuals[j];
			p_plus_k=(j+k)%NUMBER_OF_SAMPLES;
			x_t_plus_k=residuals[p_plus_k];
			num=num+(x_t-average)*(x_t_plus_k-average);
			//p++;
			//p=p%numberOfSamples;
		}
		if (num==0)
			r[k]=0; 
		else
			r[k]=num/den;
			//System.out.println("r["+k+"]="+r[k]);
	}	
		
		//analysis of the correlogram
		//we check for randomness by considering the 95% confidence limits [-2/sqrt(numberOfSamples),2/sqrt(numberOfSamples]
		//considering this confidence limits, no more than 5% of the values should be outside this range (i.e., 1 over 20)
	
		double errorPercentage=0.2;
		
		double minRange=-2.0/sqrt((double)NUMBER_OF_SAMPLES)-errorPercentage*2.0/sqrt((double)NUMBER_OF_SAMPLES);
		double maxRange=2.0/sqrt((double)NUMBER_OF_SAMPLES)+errorPercentage*2.0/sqrt((double)NUMBER_OF_SAMPLES);
		
		
		double noValuesOutOfRange=0;
		for (k=0;k<K_MAX;k++) {

			if (r[k]<1||r[k]>1) {
				
			}
			if (r[k]<minRange)
				noValuesOutOfRange=noValuesOutOfRange+1;
			if (r[k]>maxRange)
				noValuesOutOfRange=noValuesOutOfRange+1;	
		}
		
		if (noValuesOutOfRange>(20.0/100.0*(double)K_MAX+1)) {
			isPredictable=false;
		}
		
		//if (isPredictable==true)
		//	numberOfPredictableSamples=numberOfPredictableSamples+1.0;
		//numberOfValidSamples++;
	
		return isPredictable;
	}

double KalmanPredictor::getPredictability() {
		
	return (numberOfPredictableSamples/numberOfValidSamples);
}


