#include <stdio.h>
#include <string.h>
#include <math.h>
#include "omnetpp.h"
#include "host.h"
#include <fstream>
#include "mobMessage_m.h"
#ifndef SIMCONTROLLER_H
#define SIMCONTROLLER_H
#include "simController.h"
#endif /* SIMCONTROLLER_H */


using namespace std;

Define_Module(SimController);


void SimController::initialize()
{
	cModule* parent=getParentModule();
	int i;

	m_iNumberOfHosts=(int)parent->par("numberOfHosts").longValue();
	m_iNumberOfMessages=(int)parent->par("numberOfMessages").longValue();
	m_dMessageCreationInterval=parent->par("messageCreationInterval").longValue();
	m_dataCollector=(DataCollector*)(parent->getSubmodule("dataCollector"));
	
	m_iNumberOfMessagesSent = 0;
	
	m_iNumberOfBlackholeCarriers = parent->par("numberOfActiveNonCollaborativeHosts");

	if(m_iNumberOfBlackholeCarriers  > m_iNumberOfHosts)
	{
		cout<<"ERROR: bad number of blackhole nodes."<<"Aborting."<<endl;
		exit(-1);
	}

	for(i = 0; i < m_iNumberOfHosts; i++)
	{
		Host *HostModule = (Host*)(parent->getSubmodule("Hosts", i));
		HostModule->m_type = NORMAL;
	}	

	for( i = 0; i < m_iNumberOfBlackholeCarriers;i++){
		Host *HostModule = (Host*)(parent->getSubmodule("Hosts", i));
		HostModule->m_type = ACTIVE_NON_COLLABORATIVE;
		cout<<"setting node "<<i<<" as Blackhole"<<endl;
	}

	mobMessage* msg = new mobMessage("controlMsg");
	scheduleAt(simTime() + initialWaitTime, msg);

}


void SimController::handleMessage(cMessage* msg)
{
	cModule* parent=getParentModule();

	int iReceiver, iSender;
	iSender = intuniform(m_iNumberOfBlackholeCarriers+1, m_iNumberOfHosts);
	do
		iReceiver = intuniform(m_iNumberOfBlackholeCarriers+1, m_iNumberOfHosts);
	while ( iReceiver == iSender );

	Host *HostModule=(Host*)(parent->getSubmodule("Hosts",iSender-1));
	mobMessage* controlMsg = new mobMessage("controlMessage");

	controlMsg->setReceiver(iReceiver);
	controlMsg->setSender(iSender);
	controlMsg->setIntermediateRecipient(0);
	//set type 2, i.e. this is a control message
	controlMsg->setType(CONTROL_MSG);
	controlMsg->setMessageId(m_iNumberOfMessagesSent);
	// added "0" as parameter for compatibility with the
	// new interface of "sendDirect()"
	sendDirect(controlMsg,0.001,0,HostModule,"controlInput");

	m_iNumberOfMessagesSent++;
	if ( iSender < m_iNumberOfBlackholeCarriers )
	{
		m_dataCollector->m_iNumberOfBlackholeMessages++;
	}else
		m_dataCollector->m_iNumberOfGoodMessages++;

	if ( m_iNumberOfMessagesSent < m_iNumberOfMessages ) 
		scheduleAt(simTime()+m_dMessageCreationInterval,msg);

}//end handleMessage


void SimController::finish()
{
	recordScalar("Number of Messages Sent", m_iNumberOfMessagesSent);
}
